dnl Checks the OS flavour (posix, RTAI, lxrt)
dnl Defines OSAPI to be one of rtai, lxrt and posix
dnl If it is rtai and lxrt, defines RTAI_INCLUDES
dnl If it is rtai, defines LNXKERN_INCLUDES
AC_DEFUN([OROB_CHECK_OSAPI],
[
    dnl RTAI support
    # configure options
    AC_ARG_WITH(rtai,
               AC_HELP_STRING([--with-rtai=yes|no|lxrt],
                   [configure to use RTAI]),
               [opt_rtai=$withval],[opt_rtai=no])
    AC_ARG_WITH(rtai-includes,
               AC_HELP_STRING([--with-rtai-includes=DIR],
                   [RTAI headers are in DIR]),
               [opt_rtai_includes=$withval])
    AC_ARG_WITH(linux-kernel-includes,
               AC_HELP_STRING([--with-linux-kernel-includes=DIR],
                   [linux kernel headers are in DIR]),
               [opt_lnxkern_includes=$withval])

    # make related stuff
    if test "x$opt_rtai" = "xyes"; then
       OSAPI=rtai
       AC_MSG_RESULT([Configuring to use RTAI])
       AC_CHECK_RTAI_INCLUDES(RTAI_INCLUDES, $opt_rtai_includes)
       AC_CHECK_LINUXKERNEL_INCLUDES(LNXKERN_INCLUDES, $opt_lnxkern_includes)
    elif test "x$opt_rtai" = "xlxrt"; then
       OSAPI=lxrt
       AC_DEFINE(__LXRT__, [], [If we are building for LXRT])
       AC_MSG_RESULT([Enabling LXRT userspace support])
       AC_CHECK_RTAI_INCLUDES(RTAI_INCLUDES, $opt_rtai_includes)

    elif test "x$opt_rtai" = "xno"; then :
       OSAPI=posix
       AC_MSG_RESULT([Configuring to use POSIX])
    else
       AC_MSG_ERROR([invalid option --with-rtai=$opt_rtai], 2)
    fi
    AC_SUBST(OSAPI)
])

# --- RTAI & lihux kernel includes --------------------------------------
# AC_CHECK_RTAI_INCLUDES(var, path)
AC_DEFUN([AC_CHECK_RTAI_INCLUDES],
[
   AC_MSG_CHECKING([for RTAI includes])
   AC_CACHE_VAL(ac_cv_path_rtai,
    [
       IFS="${IFS= 	}"; ac_save_ifs="$IFS"; IFS=":"
        ac_tmppath="$2:/usr/realtime/include:/usr/src/rtai/include"
        for ac_dir in $ac_tmppath; do 
            test -z "$ac_dir" && ac_dir=.
            if eval test -f $ac_dir/rtai.h; then
               eval ac_cv_path_rtai="$ac_dir"
               break
            fi
       done
       IFS="$ac_save_ifs"
    ])
   $1="$ac_cv_path_rtai"
   if test -n "[$]$1"; then
      AC_MSG_RESULT([$]$1)
   else
      AC_MSG_ERROR([cannot find RTAI includes], 2)
   fi
   AC_SUBST($1)
])

AC_DEFUN([AC_CHECK_LINUXKERNEL_INCLUDES],
[
   AC_MSG_CHECKING([for linux kernel includes])
   AC_CACHE_VAL(ac_cv_path_linuxkernel,
    [
       IFS="${IFS= 	}"; ac_save_ifs="$IFS"; IFS=":"
        ac_tmppath="$2:/usr/realtime/include:/usr/src/rtai/include:/usr/src/linux"
        for ac_dir in $ac_tmppath; do 
            test -z "$ac_dir" && ac_dir=.
            if eval test -f $ac_dir/linux/kernel.h; then
               eval ac_cv_path_linuxkernel="$ac_dir"
               break
            fi
       done
       IFS="$ac_save_ifs"
    ])
   $1="$ac_cv_path_linuxkernel"
   if test -n "[$]$1"; then
      AC_MSG_RESULT([$]$1)
   else
      AC_MSG_ERROR([cannot find linux kernel includes], 2)
   fi
   AC_SUBST($1)
])


