/*
 * Copyright (c) 2004 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.
 */

/*
 * Function main aspectTest() to simply test aspectLib
 *
 */

#define MESURE_TIME

#ifdef UNIX
#include "portLib.h"
#else
#include <vxWorks.h>
#endif
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "h2mathLib.h"

#include "aspectLib.h"

extern int aspectTest();

static const ASPECT_SEG defaultCartMap[] = {
  {1, 0, 1, 1},
  {-2, 1, -2, -1},
  {-3, 1, -3, -1},
  {3, -1, 3, 20},
  {-4, 1, -4, -3},
  {2, 1, 2, 3},
  {0.5, 0.5, 0.5, 1},
  {1, 1.5, 1, 2},
  { -1, 0, -1, 1}
};

#ifdef UNIX
static char *usage = "aspectTest [x y th = %g %g %g] [dist = %g] [angleDeg = %g]  [fill.map = %s]\n";
#endif

#define NB_LOOP 1

static  ASPECT_CONFIG sensorView={0.81581126563042261,  -1.8580717425060911,  -1.3653753953369536};
static  double dist=20;
static  double angleDeg=180;
static  char defaultMap[] = "/usr/local/robots/b21r/manips/data/laas-small/maps/gs-small.map";

static ASPECT_SEG *cartMap;


static  ASPECT_VIEW_PARAMETERS viewParameters;
static  int nbseg, nbsegin, i;
static  ASPECT_SEG *aspectCart;
static  ASPECT_POLAR_SEG *aspectPolar;
static  int *inputSegId;


#ifdef UNIX

static void printUsage (ASPECT_CONFIG *sensorView, double dist, double angleDeg,
			char *map)
{
  printf (usage, sensorView->x, sensorView->y, sensorView->theta,
	  dist, angleDeg, map ? map : defaultMap);
}


int main(int argv, char **argc) 
{
  static  char *map=NULL;
  nbsegin = sizeof(defaultCartMap)/sizeof(ASPECT_SEG);

  if (argv == 1) {
    printf ("aspectTest : using default values : \n");
    printUsage(&sensorView, dist, angleDeg, map);
  }
  else if (argv == 4) {
   sensorView.x = (double)atoi(argc[1]);
   sensorView.y = (double)atoi(argc[2]);
   sensorView.theta = (double)atoi(argc[3]);
   printUsage(&sensorView, dist, angleDeg, map);
  }
  else if (argv == 5) {
   sensorView.x = (double)atoi(argc[1]);
   sensorView.y = (double)atoi(argc[2]);
   sensorView.theta = (double)atoi(argc[3]);
   dist = (double)atoi(argc[4]);
   printUsage(&sensorView, dist, angleDeg, map);
  }
  else if (argv == 6) {
   sensorView.x = (double)atoi(argc[1]);
   sensorView.y = (double)atoi(argc[2]);
   sensorView.theta = (double)atoi(argc[3]);
   dist = (double)atoi(argc[4]);
   angleDeg = (double)atoi(argc[5]);
   printUsage(&sensorView, dist, angleDeg, map);
  }
  else if (argv == 7) {
   sensorView.x = (double)atoi(argc[1]);
   sensorView.y = (double)atoi(argc[2]);
   sensorView.theta = (double)atoi(argc[3]);
   dist = (double)atoi(argc[4]);
   angleDeg = (double)atoi(argc[5]);
   printUsage(&sensorView, dist, angleDeg, map);
   cartMap = aspectReadSegkitMap(argc[6], &nbsegin);
  }
  else {
    printUsage(&sensorView, dist, angleDeg, map);
    return 0;
  }
  return aspectTest();
}
#endif

int aspectTest()
{
#ifdef MESURE_TIME
  struct timeval timebegin;
  struct timeval timeend;

  gettimeofday(&timebegin, NULL);
#endif

#ifdef VXWORKS
  nbsegin = sizeof(defaultCartMap)/sizeof(ASPECT_SEG);
#endif

  if (!cartMap)
    cartMap = defaultCartMap;

  viewParameters.keepReversedSegs = 1;
  viewParameters.maxDistView = dist;
  viewParameters.halfAngleView = DEG_TO_RAD(angleDeg/2.0);
  viewParameters.minAspectSegAngle = 0.017;

  for (i=0;i<NB_LOOP-1;i++) {

    nbseg = aspectMapAlloc(&sensorView, &viewParameters, 
			   nbsegin, cartMap, 
			   &aspectCart, &aspectPolar, &inputSegId, 0, NULL);
    free(aspectPolar);
    free(aspectCart);
    free(inputSegId);
  }

  nbseg = aspectMapAlloc(&sensorView, &viewParameters, 
			 nbsegin, cartMap, 
			 &aspectCart, &aspectPolar, &inputSegId, 0, NULL);

  printf ("%d segs \n", nbseg);

  printf ("\naspectCart:\n");
  for (i=0 ; i<nbseg; i++) {
    printf ("%d (%g, %g) (%g, %g)\n", 
	    i, 
	    (aspectCart)[i].x1, (aspectCart)[i].y1, 
	    (aspectCart)[i].x2, (aspectCart)[i].y2);
  }

#if 0
  printf ("\naspectPolar:\n");
  for (i=0 ; i<nbseg; i++) {
    printf ("%d (%g deg, %gm) (%g deg, %g m)\n", 
	    i, 
	    RAD_TO_DEG((aspectPolar)[i].t1), (aspectPolar)[i].r1, 
	    RAD_TO_DEG((aspectPolar)[i].t2), (aspectPolar)[i].r2);
  }
#endif

  free(aspectPolar);
  free(aspectCart);
  free(inputSegId);

#ifdef MESURE_TIME
  gettimeofday(&timeend, NULL);
  printf ("%5.4lf ms\n", (double)(timeend.tv_sec*1000000 + timeend.tv_usec - 
				  (timebegin.tv_sec * 1000000 + timebegin.tv_usec))/1000);
#endif

  return 1;
}
