#include <stdio.h>
#include <math.h>
#include <stdlib.h>



#include <dbliste.h>
#include <lasdef.h>
#include <las_prot.h>
#include <las_io_p.h>
#include <image_mgr_p.h>
#include <odom_funct_p.h>    /* pos_dist et pos_merge */
#include <odometer.h>
#include <odometer_p.h>

#include "aspect_p.h"

#include <lasers_robots.h>

#include "localize.h"
#include "LaserAc_p.h"
#include "match_p.h"
#include "disambiguate_p.h"
#include "map_mgr_p.h"
#include "map_unc.h"
#include "map_util.h"
#include "map_builder_p.h"
#include "print_results_p.h"

#include "aspect/aspectLib.h"

#ifndef VERBOSE
#define VERBOSE 0
#endif
#define GAUSS_99 3.0 

/*******
 Variable associated with 
 the aspect computation:
 if the minimum distance 
 between a segment and the robot
 is higher that DistMax, this 
 segment is assumed not visible.
Unit = inch.

XXX Should depend on laser caracteristics (max range)
********/
double segkitAspectDistMax = 20.0; /* 120.0; */
double segkitVisibilityAngle = M_PI;

/*********
  states that the accuracy of
  the extremum pointsw of the segments
  of the map has a standard deviation
  of 0.5 square 
  Means that the user is able
  to give those points with
  0.5 inches of accuracy in x and y

  This is changeable.
**********/
/* Sick */
#ifdef notmetric
static double unc_map = 1.0;
#else
static double unc_map = 0.05;
#endif

/* If MergeOdometry==1 merges segkit and estimated robot position */
static int MergeOdometry = 1;

/* Filters solution according to estimated position through
   probabilistic odometer.

   If  RelocFilter==1  rejects location solutions at a distance greater 
   ThresReloc from robot's given estimated position (using odometer)

   o  RelocFilter  value is changed using LocFilterOn/LocFilterOff functions 
   o  ThresReloc   value is changed using SetRelocThres function 
*/
static int RelocFilter = 0;
static double ThresReloc = 20.;


/* Tells if we have or not a probabilistic odometry
   (which is used by RelocFilter.)
   In the negative case (ie, ProbabilistFlag==0), 
   a cartesian distance is computed intead of a probabilistics one and a 
   a fixe threshold is applied.

  Remark: for robot that does not have probabilistic odometry, 
  one can used a simulated one's by setting RealVariance value to 0 */
int ProbabilistFlag = 1;

/* Prototypes of local functions */
static int RelocationFilter(PDBL *l);
static long LocalAspect(double *,double, laserConfiguration *sick_config);
static void Unset_angle_restriction(void);
static Sgl *SearchNext(Sgl *s,igla *ima);
static void LinkBoundary(void);
static Sgl *ExternSegment(igla *);
static int convertSegkitToAspectMap(igla *igabs, 
				    ASPECT_SEG **cartMap,
				    Sgl ***mapSegsId);
static int convertAspectToSegkitMap(int nbSegs, 
				    ASPECT_CONFIG *config,
				    ASPECT_SEG *aspect, 
				    int *segsId,
				    Sgl **mapSegsId,
				    long *asp);
static int printSegkitMap(igla *igabs) ;
static int localAspectAspect(double *xp,double distmax,
			     laserConfiguration *sick_config);

#ifdef UNUSED
void Set_angle_restriction(void);  
#endif /* UNUSED */

/* >>> GLOBAL VARIABLES FOR DEBUG PURPOSE (UNDER VXWORKS) */
/* For probabilistic test */
double GminDistDebug;
/* value exagerated for NOT using it unless things are very desesperate */
int ValLimFailures = -10000;
/* <<< end GLOBAL VARIABLES FOR DEBUG PURPOSE (UNDER VXWORKS) */

/******

  added 05/11/94: functions for interfacing with
  the map builder: 
  When the map is driven from outside, one must be careful not to forget
  to create an auxilliary structure "Local_map" using the function
  "Imgl_create_copy" defined in las_struct.c 
  This structure will be used by the localization function.
  I will add a function "Rinit_map" which reallocate and free (if necessary)
  this auxilliary structure. 
  
  For now, do
  

  if needed do 
  FreeLocalMap();
  SetMap(extern_map);
  extern_local_map = Imgl_create_copy(extern_map);
  SetLocalMap(extern_local_map);

  See "Map_mgr_p.h" in the include directory for
  function declarations (or track "added 05...")
**********************************/


/******** 
  Map image 
********/
static igla *Map = NULL;    
static long Map_ind;   

/******** 
  Restricted Map image 
********/
static long restricted = 0; 
static igla *RMap = NULL; 
static long RMap_ind; 

/**********
  Boundary of the map:
  means the robot can't be ouside this polygon.
***********/
static igla *Boundary = NULL; 
static long Bound_ind;

/*********
 auxilliary structure
 for expressing the map in local 
 coordinates
**********/
static igla *Local_map = NULL;

#ifdef UNUSED
#define MAP_MGR_NB_MAX_ASPECT_SEGMENTS  200
static long Aspect_segments[MAP_MGR_NB_MAX_ASPECT_SEGMENTS*2 + 1];   /* added by David, to access aspect data */
#endif /* UNUSED */

/*********
Current Data image 
*********/
static igla *Current_data = NULL;
static long Current_ind;

/********
 Local Aspect Image Id
********/
static igla *Aspect = NULL;
static long AspectId = -1;


/*******
Auxilliary (to be discarded)
for matching list between the map
and current data
********/

static PDBL *ListPrehyp = NULL;
static PDBL *ListHypo = NULL;
static PDBL *ListValidHypo = NULL;


/***** 
       For recording the number of possibilities and the success of localization 
*****/

static long NPossibilities = 0;
static long NSuccess = 0;



#define deg2rad(x) (3.141592653589793 * (x) /180.)



/* prototype the function that David added, by david */
/* All these functions are currently not in used */
#ifdef UNUSED
void get_aspect_ptr(long *aspect_ptr);
void compute_aspect (double x, double y,
		     double max_dist, long *aspect_segments, int max);
struct polar_line_segment_type *Find_Aspect (double *xp, 
					     double distmax, 
					     double a1, double a2);
void Init_Rmap(void);
void Unset_Rmap(void);
void NewRSegment(double x1,double y1,double x2,double y2);
#endif

/****
  Compute in ListValidHypo a list of valid hypothesis
  about the robot location when it acquired the data
  stored in the image n_data. This location 
  is computed according to the current map stored
  in Map_ind (index) and Map (structure itself).

  No assumption is made about the location of the robot.
  However, the user can restrict the search to a particular
  angle area by using Set_angle_restriction() described below.
  This restriction will automatically be released at the end of
  this function.

  Fills the array possloc with all the possible locations and
  returns the index on the best one in this array.
*****/

long Global_localization(long n_data,long *possloc, long op, int maxarraysize)
{
  long best;
 
  /*************
    checking on the map 
  *************/
  if(Map == NULL){
    printf("Global localization: there is no map !!!!\n");
    return(-1);
  }
 /* storing n_data in local variables */
  if(n_data <  0)
    return(-1);
  Current_ind = n_data;
  Current_data = PtoIm(n_data);
  if(Current_data == NULL)
    return(-1);

  /* freeing the previous list of hypothesis */
  /* ListValidHypo is a subset of ListHypo remaining
     after the filter test_validity_hypo() */
  
  Free_list_hyp(ListHypo);
  dlib_list(ListValidHypo);
  ListHypo = ListValidHypo = NULL;

  /* initialize the computation (match.c) */
  if (restricted) {
    if ((RMap == NULL) || (RMap_ind < 0)) {
      printf("No Restricted Map!\n");
      restricted = 0;
      RMap = Map;
      RMap_ind = Map_ind;
    }
  }
  else {

    RMap = Map;
    RMap_ind = Map_ind;    
  }

  Init_Matching(Current_data,RMap,Map);

  /* grouping the possible matchings into angular groups (match.c) */
  if((ListPrehyp =  Angular_group(Current_data,RMap)) == NULL)
    {

      NPossibilities = 0;
	  
	  /* was there success before ? */
	  if (NSuccess <= 0)
	    NSuccess -= 1;
	  else
	    NSuccess = -1;
	  
	  return(-1);
	  
	  
    }
    

  /* computes all possible hypothesis without
     taking into account occlusions (match.c) */
  if((ListHypo = Gen_Exp_Lhyp(ListPrehyp)) == NULL)
    {
       NPossibilities = 0;
	  
	  /* was there success before ? */
	  if (NSuccess <= 0)
	    NSuccess -= 1;
	  else
	    NSuccess = -1;
	  
	  return(-1);
	

    }

  
  /*saving the number of hypothesis before filtering */

  NPossibilities = Db_length(ListHypo);

  /* 
     If parameter number 5 is equal to 1:

         eliminates solutions leading to an object to be seen behind
	 a static obstacle of the map. "behind" means "further than 8. inches 
	 (last parameter) behind any static obstacle" (match.c).
	 
     Eliminates also any solution leading the robot to be
     outside the boundary "Boundary", if any.

     Filters any solution better than or equal to 2 matchings (before the last 
     parameter).
     
  */

      if((ListValidHypo = test_validity_hypos(Current_data,Map,Boundary,ListHypo,op,2,1.))
                    == NULL)


	{
	  NPossibilities = 0;
	  
	  /* was there success before ? */
	  if (NSuccess <= 0)
	    NSuccess -= 1;
	  else
	    NSuccess = -1;
	  
	  return(-1);
	  
	  
	}
  ComputeFinalListHypo(ListValidHypo);
  
  /* groups possibly compatible solution into one */
  Expurge_multiplicity(ListValidHypo);
  
  
  /*********
    computing data structures for later disambiguating (disambiguate.c)
    Fills up the array possloc with all the possible hypothesis.
    Returns also the index on the best hypothesis.
  *********/
  
  best = TransformListHyp(ListValidHypo,Map,Local_map,Boundary,
			  Current_data->enc,possloc,maxarraysize);
  

  /* looking at what is left after filtering */

  if(possloc[0] > 0){
    NPossibilities = possloc[0];
    /* was there success before ? */
    if (NSuccess >= 0)
      NSuccess += 1;
    else
      NSuccess = 1;
  }
  else{
    /* no possibilities : we mark this by inversing the number before filtering */

    NPossibilities *= -1;
    
    /* was there success before ? */
    if (NSuccess <= 0)
      NSuccess -= 1;
    else
      NSuccess = -1;

  }

  /************
      Unset any angle restriction 
  ************/
  Unset_angle_restriction();
  
  return(best);
}


#ifdef UNUSED
/****
  computing the aspect  (unused !)
*****/
void compute_aspect (double x, double y, 
		     double max_dist, long *aspect_segments, int max)
{
  double pt_view[3];
  pt_view[0] = x;
  pt_view[1] = y;
  pt_view[2] = 0.;
  AspectId = LocalAspect(pt_view,max_dist);
  Copy_image_segment (AspectId, aspect_segments, max);
}
#endif /* UNUSED */


/****
  Compute in ListValidHypo a list of valid hypothesis
  about the robot location when it acquired the data
  stored in the image n_data. This location 
  is computed according to the current map stored
  in Map_ind (index) and Map (structure itself).

  This function takes the location xl,yl as an estimate of
  the robot position. In order to work, the aspect ratio from 
  the approximate location must contain at least 2 non-parallel
  segments visible from the actual position of the robot.
  
  In addition, the user can restrict the search to a particular
  angle area by using Set_angle_restriction() described below.
  This restriction will automatically be released at the end of
  this function.

  
  if possloc non null
     Fills the array possloc with all the possible transformations enc->position


     NOTE : the hypothesis are available  as long as a localization function is not called again.
     Maybe it would be better to separately call the localization function and then bring 
     back the result in the form of transformation or absolute position.

*****/
long Local_localization(long n_data,double xl,double yl,long *possloc,
			long op, int maxarraysize, 
			laserConfiguration *sick_config)
{
  double pt_view[4];
  long best;
#if VERBOSE > 4
  Hyp *Best;
  PDBL *possibleTransfs;
#endif
  int nbmatchings;
  double mean_angle, gap_angle; /* Mean and gap init angle */

  /*************
    checking on the map 
  *************/
  if(Map == NULL){
    printf("Local_localization: there's no map !!!!\n");
    return(-1);
  }
  /*******
    storing n_data
  ********/
  Current_ind = n_data;
  Current_data = PtoIm(n_data);
  if(Current_data == NULL) {
    printf("Local_localization: there's no data !!\n");
    return(-1);
  }

  /***********
    freeing previous strutures
  ***********/

  Free_list_hyp(ListHypo);
  dlib_list(ListValidHypo);

  ListValidHypo = ListHypo = NULL;
  /****
    freeing the aspect image
  ****/
  Gfree_image(AspectId);

  /* checking if the local map exists */

#if VERBOSE > 2
  printf ("--- Init pos  : %.3f %.3f %.3f\n", 
	  Glob_pos[0], 
	  Glob_pos[1], 
	  Glob_pos[2]);

  printf ("   Init Vpos : %.3f %.3f %.3f\n", 
	  sqrt(Vpos[0]) * GAUSS_99,
	  sqrt(Vpos[2]) * GAUSS_99,
	  sqrt(Vpos[5]) * GAUSS_99);
#endif

  if(! Local_map)
    Local_map = Imgl_create_copy(Map);

  /**************************************************************
   ** Computing the ASPECT 
   **  (ie, the segments of the map that are visible from the 
   ** estimated robot position (xl, yl, ) 
   **************************************************************/

  pt_view[0] = xl;
  pt_view[1] = yl;
  /* XXXXXXXX  pt_view[2] = 0.; --->> WHY ? We have data on the angle ! 
     (however, why those data are not parameters of Local_localization ?) */
  Get_loc_angle(&mean_angle, &gap_angle);
  pt_view[2] = mean_angle;
  /* And I use pt_view[3] to give gap_angle !!! */
  pt_view[3] = gap_angle;

  /* Compute the Aspect */
  AspectId = LocalAspect(pt_view, segkitAspectDistMax, sick_config); 
/*   AspectId = localAspectAspect(pt_view, segkitAspectDistMax, sick_config); */


  /* Get the pointer on the aspect map */
  Aspect = PtoIm(AspectId);
  if (Aspect->nsg <= 0) {
    printf("Local_localization: no segs in aspect map\n");
    return(-1);
  }
#if VERBOSE>0
  printf ("-- LocalAspect -> %ld segs\n", Aspect->nsg);
#endif
#if VERBOSE>3
  printSegkitMap(Aspect);
#endif
  
#ifdef UNUSED
  /* Just a copy in an long array */
  Copy_image_segment (AspectId, Aspect_segments, 
		      MAP_MGR_NB_MAX_ASPECT_SEGMENTS);
#endif


  /*****************************************************************
   **  Initializing the matching process 
   *****************************************************************/
  /* Filters percept segs (they are classified according to angle 
     confidance and small segs are removed) and
     pre-computes an array "match" (in match.c) of all 
     possible combination between perception data 
     and the map. */
  nbmatchings = Init_Matching(Current_data,Aspect,Map);
#if VERBOSE>0
  printf ("-- Init_Matching -> %d matchings\n", nbmatchings);
#endif

  /*****************************************************************
   **  grouping into angular groups
   *****************************************************************/
  /* Computes a list of pre-hypothesis of matching. All elements in a 
     pre-hyp are compatible in angle according to a Mahalanobis distance test.
     It is  a partitionning : one matched pair  belongs to one pre-hyp. 

     If flag LocAng==1 then groups which angle is out of 
     'MeanAngle +/- GapAngle' interval are not considered.
     (see Set_loc_angle and Unset_loc_angle).
*/
  if((ListPrehyp =  Angular_group(Current_data,Aspect)) == NULL)
    	{
	  NPossibilities = 0;
	  NSuccess = 0;
	  /* was there success before ? */
	  /*if (NSuccess <= 0)
	    NSuccess -= 1;
	  else
	    NSuccess = -1;
	    */	  
	  printf("Local_localization: no valid angle pre-hyp\n");
	  return(-1);
	}
    
#if VERBOSE>0
    printf ("-- Angular_group -> %ld groups\n", Db_length(ListPrehyp));
#endif 
#if VERBOSE>4
    print_pre_hyp_list(stdout, ListPrehyp,  0);
#endif

  /****************************************************************
   ** computes all possible hypothesis without
   ** taking into account occlusions (match.c) 
   ****************************************************************/

  if((ListHypo = Gen_Exp_Lhyp(ListPrehyp)) == NULL)
    
  	{
	  NPossibilities = 0;
	  NSuccess = 0;
	  /* was there success before ? */
	  /*if (NSuccess <= 0)
	    NSuccess -= 1;
	  else
	    NSuccess = -1;
	    */	  
	  printf("Local_localization: no hyp found\n");
	  return(-1);
	}

  /*saving the number of hypothesis before filtering */

  NPossibilities = Db_length(ListHypo);
#if VERBOSE>0
  printf ("-- Gen_Exp_Lhyp -> %ld hyps\n", NPossibilities);
#endif 
#if VERBOSE>4
    print_hyp_list(stdout, ListHypo, NULL, 1, 0, 0);
#endif

  /* 
     If parameter number 5 is equal to 1:

         eliminates solutions leading to an object to be seen behind
	 a static obstacle of the map. "behind" means "further than 20 centimeters 
	 (last parameter) behind any static obstacle" (match.c).
	 
     Eliminates also any solution leading the robot to be
     outside the boundary "Boundary", if any.

     Filters any solution better than or equal to 2 matchings (before the last 
     parameter).
     
  */

  
  if((ListValidHypo = test_validity_hypos(Current_data,Map,Boundary,ListHypo,op,2,0.20))
                    == NULL)
   
  	{
	  NPossibilities = 0;
	  NSuccess = 0;
	  /* was there success before ? */
	  /*  if (NSuccess <= 0)
	    NSuccess -= 1;
	  else
	    NSuccess = -1;
	    */
	  printf("Local_localization: no valid hyp\n");
	  return(-1); 
	 
	  
	}
#if VERBOSE>0
    printf ("-- test_validity_hypos -> %ld hyps\n", Db_length(ListValidHypo));
#endif 
  ComputeFinalListHypo(ListValidHypo);
#if VERBOSE>0
    printf ("--ComputeFinalListHypo -> %ld hyps\n\n", Db_length(ListValidHypo));
#endif 
#if VERBOSE>4
    print_hyp_list(stdout, ListValidHypo, NULL, 1, 0, 0);
#endif

  /* groups possibly compatible solution into one */
  /*Expurge_multiplicity(ListValidHypo);*/

  

  /** Added 6/26/94 : if the option RelocFilter is on, 
    we test if the distance between the locations found and the estimate
    location. Warning: relies  on the values Glob_pos and Vpos which are supposed 
    to be compatible with the current estimate position.
    Not to be used if those values are not updated.
    ***/
  if(NSuccess > ValLimFailures){
      if(RelocFilter) {
	int nbRelocHyp;
	if ((nbRelocHyp = RelocationFilter(ListValidHypo)) == 0) {
#if VERBOSE>0
	  printf("Local_localization: no pos-compatible hyp\n");
#endif
	}
#if VERBOSE>0
	printf ("-- RelocationFilter -> %d hyps\n", nbRelocHyp);
#endif 
      }
  }

  /*********
    computing data structures for later disambiguating (disambiguate.c)
    Fills up the array possloc with all the possible hypothesis.
    (Computes length_matched.)
    Initializes the robot position and variance from the best hypothesis and
    returns the index on the best hypothesis.
*********/
  
  best = TransformListHyp(ListValidHypo,Map,Local_map,Boundary,
			  Current_data->enc,possloc,maxarraysize);


#if VERBOSE>0
	printf ("-- TransformListHyp -> %ld hyps\n", possloc[0]);
	printf ("   TransformListHyp : best %ld \n", best);
#endif 
#if VERBOSE>4
	possibleTransfs = GetPossibleTranfs();
	Best = GetBestHyp();
	print_hyp_list(stdout, ListValidHypo, Best, 1, 1, 0);
#endif

  /* looking at what is left after filtering */

  if(possloc[0] > 0){
    NPossibilities = possloc[0];
    /* was there success before ? */
    if (NSuccess >= 0)
      NSuccess += 1;
    else
      NSuccess = 1;
  }
  else{
    /* no possibilities : we mark this by inversing the number before filtering */

#if VERBOSE>0
    printf ("? ");
    fflush(stdout);
#endif

    NPossibilities *= -1;
    
    /* was there success before ? */
    if (NSuccess <= 0) {
      NSuccess -= 1;
    }
    else
      NSuccess = -1;

  }

  
  /************
    Unset any angle restriction 
  ************/

  Unset_angle_restriction();

 
  return(best);
}

/**** normally the localization produces an array of transformations ***/
/**** 
      however, the list of positions corresponding to the position from   which 
      the data was taken  is still available so here is the functiuon to access it 
****/

long GetPositionArray(long *pos, int max)
{
  PDBL *l;
  Hyp *hyp;
  int n = 0;

  
  
  for(l = ListValidHypo; l != NULL; l = dls(l)){
    hyp = (Hyp *)dlel(l);

    /*******
      I added a field called "->length_matched in which
      the total length matched is computed by this function. To be done
    *******/
    /* Compute_length_matched(hyp);*/
    if(hyp->nmatched >= 2)
      {
	/*****
	  copying into the array possloc
	  WARNING: the array has to be big enough !!!
	  = 4*number of possible location + 1
	******/

	if(n < max) {
	  pos[4*n + 1] = hyp->Tr[0]*1000;
	  pos[4*n + 2] = hyp->Tr[1]*1000;
	  pos[4*n + 3] = hyp->Tr[2]*1000;
	  pos[4*n + 4] = hyp->nmatched;
	  
	 
	}
	else {
	  printf("GetPositionArray Warning: pos array too small only %d pos recorded\n",  n);
	  break;
	}
	n++;

	
      }
  }
  
    
  
  /*** returning the result *****/
  
  pos[0] = n;

  return(n);
  


}
/*Well, this is dealing only with the position, not the angle ... 
 * If we want to introduce the current position value, it should come here .
 * Here it comes : if the variable MergeOdometry is defined,  hyp->Tr and Hyp->Vtr are recomputed
 * according to Glob_pos and Vpos
 */



static int RelocationFilter(PDBL *lhyp)
{
  PDBL *l;
  Hyp *hyp;
  double dist,ecx,ecy;
  double vx,vy;
  static double vdummy[6];
  double min_dist = 100000000.;
  int nbValidPos=0;

  for(l = lhyp; l != NULL; l = dls(l)){
    hyp = (Hyp *)dlel(l);
   

    if(! ProbabilistFlag){    
      ecx = Glob_pos[0] - hyp->Tr[0]; 
      ecx = ecx*ecx; 
      ecy = Glob_pos[1] - hyp->Tr[1]; 
      ecy = ecy * ecy; 
      
      if(Vpos[0] < 0.0004) 
	vx = 0.0004; 
      else 
	vx = Vpos[0]; 
      if(Vpos[2] < 0.0004) 
	vy = 0.0004; 
      else 
	vy = Vpos[2]; 
      if (ecx > vx*ThresReloc || ecy > vy*ThresReloc)
	
	hyp->nmatched = 0;
      
      /* yes : merging if option set */
      else  {
	nbValidPos++;
	if (MergeOdometry){
	
	  pos_merge(Glob_pos,Vpos,hyp->Tr,hyp->Vtr,hyp->Tr,vdummy,0);
	}
      }
    }
    else {
      /*testing the likelyhood of Glob_pos == hyp->Tr */
      dist =  pos_dist(Glob_pos,Vpos,hyp->Tr,hyp->Vtr,hyp->Tr,vdummy);

      if(dist < min_dist && dist > 0)
	min_dist = dist;

      /* Ok ??? ... */
      if( dist > ThresReloc || dist < 0) {
	if (dist < 0)
	  printf("dist < 0 !\n");
#if VERBOSE>2
	printf("   dist (%.3f %.3f %.3f) = %6.1f -> NO  nm %ld lm %g\n",
	       hyp->Tr[0], hyp->Tr[1], hyp->Tr[2], dist, 
	       hyp->nmatched, hyp->length_matched);
#endif

	/* no : discarding */
	
	
	hyp->nmatched = 0;
      }

      /* yes : merging if option set */
      else  {

	if (!pos_jump_filter(Glob_pos,Vpos,hyp->Tr)) {
	  hyp->nmatched = 0;
	}
	else  {

	  nbValidPos++;

#if VERBOSE>4
	printf("OK dist (%.3f %.3f %.3f) = %6.1f  nm %ld\n",
	       hyp->Tr[0], hyp->Tr[1], hyp->Tr[2], dist, 
	       hyp->nmatched);
#endif
	if (MergeOdometry){
	  /* Merge solution with predicted pos */
	  /* XXXX The variance must also be merged ... */
/* 	  pos_merge(Glob_pos,Vpos,hyp->Tr,hyp->Vtr,hyp->Tr,vdummy,0); */
	  pos_merge(Glob_pos,Vpos,hyp->Tr,hyp->Vtr,hyp->Tr,hyp->Vtr,0);
#if VERBOSE>1
	  printf("ESTIM   (%.3f %.3f %.3f)\n",
		 Glob_pos[0], Glob_pos[1], Glob_pos[2]);
	  printf("merge   (%.3f %.3f %.3f)\n",
		 hyp->Tr[0], hyp->Tr[1], hyp->Tr[2]);
#endif
	}}
      }
    }
  }
  GminDistDebug = min_dist;
  return nbValidPos;
}

void SetRelocThres(long n){
  ThresReloc = (double)n;
}

void LocFilterOn()
{RelocFilter = 1;}
void LocFilterOff()
{RelocFilter = 0;}

void MergeOdoOn()
{MergeOdometry = 1;}
void MergeOdoOff()
{MergeOdometry = 0;}



/*** accessing the records ****/

long GetNpossibilities ()
{
  return NPossibilities;

}

long GetNSuccess ()

{
  return NSuccess;
}


#ifdef UNUSED
/***********
 let user code to access Aspect_segments
 added by David: 1/20/94
************/
void get_aspect_ptr (long *aspect_ptr)
{
  long i;
  long nbSegs;

  nbSegs = Aspect_segments[0];
  if (nbSegs > MAP_MGR_NB_MAX_ASPECT_SEGMENTS) {
    printf ("get_aspect_ptr: too many segments (%d recorded instead of %d)\n",
	    MAP_MGR_NB_MAX_ASPECT_SEGMENTS, nbSegs);
    nbSegs=MAP_MGR_NB_MAX_ASPECT_SEGMENTS;
  }
  for (i=0; i<=4*nbSegs; i++) {
    aspect_ptr[i] = Aspect_segments[i];
  }
}
#endif /* UNUSED */



/***********
  computes the local aspect of the Map from the position xp
  eliminates parts farther than distmax from xp.
  returns an index on the aspect image created.

  Warning: the user has to free the image afterwards (use Gfree_image(index)).
  returns -1 if there is no map
************/
#define LOCAL_VISI_FAST_APPROX   0
#define LOCAL_VISI_EXACT_FILTER  1
#define LOCAL_VISI_ORIGINAL      2
/* #define DEFAULT_FAST_ASPECT */
#ifdef DEFAULT_FAST_ASPECT
int local_visibility_method = LOCAL_VISI_FAST_APPROX;
#else
int local_visibility_method = LOCAL_VISI_EXACT_FILTER;
#endif


/* ----------------------------------------------------------------------
 *
 * localAspectAspect
 *
 */

static int localAspectAspect(double *xp,double distmax,  
			     laserConfiguration *sick_config)
{
  ASPECT_CONFIG config = {xp[0], xp[1], xp[2]};
  ASPECT_VIEW_PARAMETERS viewParameters;
  int nbSegs;
  ASPECT_SEG *aspectCartToAlloc;
  int *outputSegId;
  int nbAspectSegs;
  long aspectId;

  static ASPECT_SEG *cartMap=NULL;
  static Sgl **mapSegsId=NULL;

  /* --- 
     --- Sensor configuration
     --- 
  */

  /* Position of the sensor within the map */
  config.x = xp[0] + sick_config->pos[0]*cos(xp[2]) - sick_config->pos[1]*sin(xp[2]);
  config.y = xp[1] + sick_config->pos[0]*sin(xp[2]) + sick_config->pos[1]*cos(xp[2]);
  config.theta = xp[2] + sick_config->pos[2];

  /* view parameters */
  viewParameters.keepReversedSegs = 0;
  viewParameters.maxDistView = distmax;
  /* Angle within segments can been seen by sick, i.e.,
     angle of view of the sick + angle uncertainty.
     (I put this last gap in x[3]) */
  viewParameters.halfAngleView  = segkitVisibilityAngle/2. + xp[3];
  viewParameters.halfAngleView  = 0;
  /* minimum angle within a seg is view (to filter far little segs) */
  viewParameters.minAspectSegAngle = DEG_TO_RAD(1.5);


  /* --- 
     --- Convert map 
     --- 
  */
  nbSegs = convertSegkitToAspectMap(Map, &cartMap, &mapSegsId);


  /* --- 
     --- Compute aspect
     --- 
  */
  nbAspectSegs = aspectMapAlloc(&config, &viewParameters, nbSegs, cartMap,
				1, /* convert back in globa frame */
				&aspectCartToAlloc, NULL, 
				&outputSegId,0, NULL);

  /* --- 
     --- Convert back to segkit format
     --- 
  */
  convertAspectToSegkitMap(nbAspectSegs, &config, aspectCartToAlloc, 
			   outputSegId, mapSegsId, &aspectId);

  /* --- 
     --- End
     --- 
  */

  free(aspectCartToAlloc);
  free(outputSegId);

  return aspectId;
}

static int convertSegkitToAspectMap(igla *igabs, 
				    ASPECT_SEG **cartMap,
				    Sgl ***mapSegsId)
{
  PDBL *la;
  Sgl *sa;
  int i=0;
  static igla *prevIgabs=NULL;
  static ASPECT_SEG *prevCartMap=NULL;
  static Sgl **prevMapSegsId=NULL;

  /* test if usefull */
  if (igabs == prevIgabs &&
      *cartMap == prevCartMap &&
      *mapSegsId == prevMapSegsId)
    return igabs->nsg;

  /* alloc aspect map */
  if (*cartMap)
    free(*cartMap);
  if (!(*cartMap = malloc (igabs->nsg * sizeof(ASPECT_SEG)))) {
    printf ("convertSegkitToAspectMap : cannot alloc %ld segs\n", 
	    igabs->nsg);
    return 0;
  }

  /* alloc array to record map segs addresses */
  if (*mapSegsId)
    free(*mapSegsId);
  if(!(*mapSegsId = malloc(igabs->nsg * sizeof(Sgl *)))) {
    return 0;
  }

  /* record map and map segs addresses */
  for(la = igabs->lsgl; la != NULL; la = dls(la)){
    sa = (Sgl *)dlel(la);
    (*mapSegsId)[i] = sa;
    (*cartMap)[i].x1 = sa->first.x;
    (*cartMap)[i].y1 = sa->first.y;
    (*cartMap)[i].x2 = sa->last.x;
    (*cartMap)[i].y2 = sa->last.y;
    i++;
  }

  /* recall old adresses */
  prevIgabs = igabs;
  prevCartMap = *cartMap;
  prevMapSegsId = *mapSegsId;

  return i;
}

static int printSegkitMap(igla *igabs) 
{
  PDBL *la;
  Sgl *sa;
  int i=0;

  wrt_file_sg_c(igabs, "/home/sara/res");

  for(la = igabs->lsgl; la != NULL; la = dls(la)){
    sa = (Sgl *)dlel(la);
    printf ("%3d: (%5.3f %5.3f) (%5.3f %5.3f)\n",
	    i++,
	    sa->first.x,
	    sa->first.y,
	    sa->last.x,
	    sa->last.y);
  }
  return i;
}

static int convertAspectToSegkitMap(int nbSegs, 
				    ASPECT_CONFIG *config,
				    ASPECT_SEG *aspect, 
				    int *segsId,
				    Sgl **mapSegsId,
				    long *asp)
{
  int i;
  Sgl *sg;
  PDBL *l2;
  igla *aspectl;

  /* init output aspect map */
  *asp = Get_Image();
  aspectl = PtoIm(*asp);
  Set_pos_image(aspectl, config->x, config->y, config->theta, config->theta);
  aspectl->coosys = 'S';

  /* fill in aspect map */
  for (i=0; i<nbSegs; i++) {
    
    if((sg = Sgl_cree()) == NULL) {
      return 0;
    }
    /* copy original map seg */
    *sg = *(mapSegsId[segsId[i]]);

#if VERBOSE>4
    printf ("    map %3d (%5.2f %5.2f) (%5.2f %5.2f)\n"
	    " -> asp %3d (%5.2f %5.2f) (%5.2f %5.2f)\n",
	    segsId[i], sg->first.x, sg->first.y, sg->last.x, sg->last.y,
	    i, aspect[i].x1, aspect[i].y1, aspect[i].x2, aspect[i].y2);
#endif

    /* record new coord */
    sg->first.x = aspect[i].x1;
    sg->first.y = aspect[i].y1;
    sg->last.x = aspect[i].x2;
    sg->last.y = aspect[i].y2;
    sg->ifirst = segsId[i];
    if(aspectl->lsgl == NULL){
      aspectl->lsgl = dlbcons(NULL,(char *)sg);
      l2 = aspectl->lsgl;
    }
    else
      l2 = dlbcons(l2,(char *)sg);
  }
  aspectl->nsg = nbSegs;

  return nbSegs;
}

static long LocalAspect(double *xp,double distmax,
			laserConfiguration *sick_config)
{
  igla *aspect, *ig;
  long asp;
  double visibility_angle_max;
#define  TEST_FREE_SPACE 1


  if(Map == NULL){
    printf("LocalAspect: there is no map !!!\n");
    return(-1);
  }

  /**
   **  Projecting the map in the local coord. sys 
   **/
  xp[0] += sick_config->pos[0] * cos(xp[2]) - sick_config->pos[1] * sin(xp[2]);
  xp[1] += sick_config->pos[0] * sin(xp[2]) + sick_config->pos[1] * cos(xp[2]);
  xp[2] += sick_config->pos[2];
  tr_abs_to_loc(Map, Local_map, xp,'S');
  
  /* Initializing aspect image */
  asp = Get_Image();

  aspect = PtoIm(asp);

  /* Initializing intermediate image ig 
     (the xp coord. are not used in local_visibility functions,
     but in following Proj_ima_s */
  ig = Imgl_cree();

  Set_pos_image(ig,xp[0],xp[1],xp[2],xp[2]);

  ig->coosys = 'S';

  /** 
   ** Computing the local aspect 
   ** ie, the segments visible from the (xp[0],xp[1],xp[2]) vu point 
   ***/
  
  /* Angle within segments can been seen by sick, i.e.,
     angle of view of the sick + angle uncertainty.
     (I put this last gap in x[3]) */
  visibility_angle_max = segkitVisibilityAngle/2. + xp[3];

  /* Original method */
  if (local_visibility_method == LOCAL_VISI_ORIGINAL)
    Local_visibility(Local_map, ig, distmax);
  /* Similar to original method but faster because the distant segments and 
     the segments "behind" the robot are filter previously to the real 
     visibility computation (it is not recommended to remove reverse 
     segments) */
  /* XXX Would be nice to take into account position uncertainty
     XXX (by translating it to an angular uncertainty according to 
     XXX  the distances robot-seg1-seg2) */
  else if (local_visibility_method == LOCAL_VISI_EXACT_FILTER)

/* #define ASPECT_UNCERTAIN */
#ifdef ASPECT_UNCERTAIN
    {
#ifndef GAUSS_99
#define GAUSS_99 3
#endif

    double pos_uncertainty; /* Incertitude sur la position en metres */

    pos_uncertainty = GAUSS_99*sqrt(Vpos[0]+Vpos[2]);

    Local_visibility_filter_uncertain(Local_map, ig, distmax,
				      pos_uncertainty,
				      visibility_angle_max);
    }
#else

    Local_visibility_filter(Local_map, ig, distmax,
			    visibility_angle_max);
#endif

  /* Not a real visibility map but very efficent 
     and less dependant on the init pos reliability.
     However, keep more segments than the previous methods 
     (not very confident) */
  else
    Local_visibility_simple(Local_map, ig, distmax,
			    visibility_angle_max, 
#ifndef DEFAULT_FAST_ASPECT
			    TEST_FREE_SPACE);
#else
                            !TEST_FREE_SPACE);
#endif

  /**
   **  projecting back to the global coord. sys 
   **/
  Proj_ima_s(ig,aspect);

  Imgl_free(ig);
  
  return(asp);
}

#ifdef UNUSED
/***********
  computes the local aspect of the Map from the position xp
  eliminates parts farther than distmax from xp.
  returns an index on the aspect image created.

  Warning: the user has to free the image afterwards (use Gfree_image(index)).
  returns -1 if there is no map
************/
struct polar_line_segment_type *Find_Aspect
       (double *xp, double distmax, double a1, double a2)
{
  long asp;
  igla *aspect;

  if(Map == NULL){
    printf("aspect computation: there is no map !!!\n");
    return(0);
  }
  /* projecting the map in the local coord. sys */
  
  tr_abs_to_loc(Map, Local_map, xp, 'S');

  asp = Get_Image();

  aspect = PtoIm(asp);

  
  return(Find_visible_segments(Local_map, aspect, distmax, a1, a2));

}
#endif /* UNUSED */

/******************************
  Functions for creating or reading a map 
  
******************************/


/******
  Initialize the internal map structure
  Must be called each time one
  want to create a new map

*******/

void Init_map(void)
{

   
  /****** freeing the possible previous map ****/
  
  if(Map_ind > 0 && Map !=  NULL) 
    /**** checking everything is ok ***/ 
    if(Map == PtoIm(Map_ind)){
      Gfree_image(Map_ind);
    }
  
  Map_ind = Get_Image();
  Map = PtoIm(Map_ind);
}

/*** 
  added 05/10/94 : function for accessing the map from
  outside the Map Manager.
****/

long Get_map_num()
{
  return(Map_ind);
}

/*** 
  added 05/21/94 : function for accessing the map from
  outside the Map Manager.
****/

long Get_aspect_num(void)
{
  return(AspectId);
}


#ifdef UNUSED
/******
  Initialize the internal map structure
  Must be called each time one
  want to create a new map

*******/

void Init_Rmap(void)
{

   
  /****** freeing the possible previous map ****/
  
  if(RMap_ind > 0 && RMap !=  NULL) 
    /**** checking everything is ok ***/ 
    if(RMap == PtoIm(RMap_ind)){
      if(restricted)
	Gfree_image(RMap_ind);
    }
  restricted = 1;
  RMap_ind = Get_Image();
  RMap = PtoIm(RMap_ind);
}
#endif /* UNUSED */


#ifdef UNUSED
void Unset_Rmap(void)
{
  restricted = 0;
   
  /****** freeing the possible previous map ****/
  
  if(RMap_ind > 0 && RMap !=  NULL) 
    /**** checking everything is ok ***/ 
    if(RMap == PtoIm(RMap_ind)){
	Gfree_image(RMap_ind);
    }
}
#endif /* UNUSED */



/******
  Initialize the internal boundary structure.
  See ReadMap for a complete example.
*******/

void Init_boundary(void)
{
  

    /****** freeing the possible previous map ****/

  if( Bound_ind > 0 && Boundary !=  NULL) 
    /**** checking everything is ok ***/ 
    if(Boundary == PtoIm(Bound_ind)){
      Gfree_image(Bound_ind);
    }

  Bound_ind = Get_Image();
  Boundary = PtoIm(Bound_ind);
}


/***************
  adds a segment to the map;
  Segments are understood in the server map 
  format: the free space is on the right side
  when walking from x1,y1 to x2,y2 and the unit is tenth of inches.
  
  
  See ReadMap() below for an example.
****************/

void NewSegment(double x1,double y1,double x2,double y2)
{
  Sgl *s;

  s = Sgl_cree();

  
  /* we reverse for phil's format */

  s->first.x = x2/10.; s->first.y = y2/10.;
  s->last.x = x1/10.; s->last.y = y1/10.;

  /* 
     Creating the variance matrix.
     The uncertainty is assumed uniform and
     with no correlation between x and y.
     Can be replaced by a more sophisticated function if needed.
     So far, works well.
   */

  s->first.vx = s->first.vy = unc_map*unc_map/12.;
  s->first.vxy = 0.;
  s->last.vx = s->last.vy = unc_map*unc_map/12.;
  s->last.vxy = 0.;
  
  /** 
    initialization of the line parameters (las_unc.c) 
    and their variance.
  **/
 
  initsg_map(&(s->first),&(s->last),s,unc_map);
 
  /* adding to the Map */

  Map->lsgl = dlhcons(Map->lsgl,s);

  Map->nsg ++;

}


#ifdef UNUSED
/***************
  adds a segment to the map;
  Segments are understood in the server map 
  format: the free space is on the right side
  when walking from x1,y1 to x2,y2 and the unit is tenth of inches.
  
  
  See ReadMap() below for an example.
****************/

void NewRSegment(double x1,double y1,double x2,double y2)
{
  Sgl *s;

  s = Sgl_cree();

  
  /* we reverse for phil's format */

  s->first.x = x2/10.; s->first.y = y2/10.;
  s->last.x = x1/10.; s->last.y = y1/10.;

  /* 
     Creating the variance matrix.
     The uncertainty is assumed uniform and
     with no correlation between x and y.
     Can be replaced by a more sophisticated function if needed.
     So far, works well.
   */

  s->first.vx = s->first.vy = unc_map*unc_map/12.;
  s->first.vxy = 0.;
  s->last.vx = s->last.vy = unc_map*unc_map/12.;
  s->last.vxy = 0.;
  
  /** 
    initialization of the line parameters (las_unc.c) 
    and their variance.
  **/
 
  initsg_map(&(s->first),&(s->last),s,unc_map);
 
  /* adding to the Map */

  RMap->lsgl = dlhcons(RMap->lsgl,s);

  RMap->nsg ++;

}
#endif /* UNUSED */

/*********
  adding a segment to the boundary
*********/

void NewSegmentBound(double x1,double y1,double x2,double y2)
{
  Sgl *s;

  s = Sgl_cree();

  
  /* we reverse */

  s->first.x = x2/10.; s->first.y = y2/10.;
  s->last.x = x1/10.; s->last.y = y1/10.;

  sg_calc_param(s);

  /* adding to the Boundary */

  Boundary->lsgl = dlhcons(Boundary->lsgl,s);

  Boundary->nsg ++;

}


/******
  added 05/11/94: functions for interfacing with
  the map builder: set the current map from the builder;
******/

void SetMap(long map)
{
  Map = PtoIm(map);
  Map_ind = map;
}

void UnsetMap(void)
{
  Map_ind = -1;
  Map = NULL;
}

/**** 
  added 05/11/94 : setting Local_map to the auxilliary structure
  defined by a external user
****/

void SetLocalMap(igla *loc_aux)
{
  Local_map = loc_aux;
}

/*** saving map created by the map builder **/

void SaveMap(void)
{
  
  write_file_sg(Map_ind);
}


/*** saving map created by the map builder **/

long SaveMapToFile(const char *file)
{
  
  return wrt_file_sg_c(Map,file);
}

/**** 
  reading a map in the map builder internal format.
  Returns the index on the map image  
 ****/

long ReadBuiltMap(const char *file)
{
  int m;
  if( (m = read_file_sg(file)) <0)
    return -1;
  Map_ind = m;
  Map = PtoIm(Map_ind);
  Init_boundary();
  CloseMap();
  return(Map_ind);
}

/****
  added 05/11/94: freeing the auxilliary structure used for
  the computation of the aspect and the disambiguating.
****/

void FreeLocalMap()
{
  if(Local_map == NULL)
    return;

  /** freeing the user field **/
  free(Local_map->usr);
  /** freeing the image **/
  Imgl_free(Local_map);
  Local_map = NULL;
}


/*****************
  Reads a file of segments in the server format (tenth of inches and
  free space on the right) describing:
  
  - a map: set of segments connected or not describing the 2d shape
  of the robot environment. The described objects are assumed to
  be STATIC (walls or objects which don't move). They also have to not
  be see-through. This is important. That doesn't mean your environment
  has to be static. The process will take unexpected occlusions into account.

  -a boundary (optional)
  
  This is different form the map in the sens
  that it is not used for generating hypothesis
  but rather to restrict the area of the possible locations.
  Consequently, the boundary can contain "virtual" walls
  which could not be visible. It is a good 
  way to define doors or windows which could be open or
  closed but delimiting an area outside of which the robot cannot be.
  (outside doors and windows of a building, for example).

  The boundary can contain segments from the map. It is assumed
  to be a closed polygon. You don't need to repeat the possible segments
  of the map into the boundary. It needs only the extra segments necessary 
  to close the polygon. Consequently, a good way to define map + boundary is:

  - enter the entire map using NewSegment() described below.

  - enter all the extra segments for delimiting the boundary using
    NewSegmentBound() described below.

  - call LinkBoundary() for computing the actual closed polygon.

  NOTE: if LinkBoundary() can't find a closed polygon, the boundary
  information will be igored. If there is no extra closing segments,
  the function LinkBoundary() will try to find an external closed
  polygon out of the segments of the map. Thus, a closed map will
  automatically define a boundary.

  You don't need a boundary to perform the localization, 
  neither the disambiguating. However, without boundary, the localization
  process will generate more possible locations and the disambiguating
  will take a noticeable extra time to perform its job.

  
  The format of the file is assumed to be:

  - an integer for the number of the map segments
  - a list of segments (in double) x1,y1,x2,y2.
  - an integer for the number of extra closing segments (if any)
  - a list of segments (in double) x1,y1,x2,y2.
******************/

void ReadMap(char *file)
{
  FILE *ptr;
  long nsg,i;
  double x1,y1,x2,y2;
  
  /*******
    Initializing the Map
  *******/
  Init_map();
  
  /*******
    Initializing the Boundary
  *******/
  
  Init_boundary();


      
  if((ptr = fopen(file,"r")) == NULL)
    {
      printf("Can't open %s\n",file);
      return;
    }  
  /***********
    reading map segments
  ************/

  fscanf(ptr,"%ld",&nsg);
  for(i = 0; i < nsg; i++){
    fscanf(ptr,"%lf %lf",&x1,&y1);
    fscanf(ptr,"%lf %lf",&x2,&y2);
    NewSegment(x1,y1,x2,y2);
    
  }
  
    
  /***********
    reading closing segments for the boundary
  ***********/

  if(fscanf(ptr,"%ld",&nsg) != EOF){
    for(i = 0; i < nsg; i++){
     
      fscanf(ptr,"%lf %lf",&x1,&y1);
      fscanf(ptr,"%lf %lf",&x2,&y2);
      NewSegmentBound(x1,y1,x2,y2);
      
    }
  }
  fclose(ptr);

  CloseMap();
 
}

void CloseMap(void){
 
  /********
    links the segments of the boundary which 
    must be a closed polygon
  ********/

  LinkBoundary();

  /*** creating a copy of the map for projecting 
       it into local coordinate system.
  ***/

  Local_map = Imgl_create_copy(Map);

  printf("Map number %ld initialized\n",Map_ind);
  printf("%ld segments in the map \n",Map->nsg);
  Export2MapBuilder(Map_ind);

}


void SetBoundary(long bound)
{
  if (bound < 0)
    Boundary = NULL;

  else
    Boundary = PtoIm(bound);
}


static void LinkBoundary(void)
{
  Sgl *Current,*Previous,*Begin,*s,*sp,*sbeg;
  long ClosingSegments = 1;


  /******* no extra closing segments ******/
  if(Boundary->nsg == 0){

    /* looking if there exists a closed boundary in
       the map itself
    */

    ClosingSegments = 0;
    Begin = ExternSegment(Map);
    s = Sgl_cree();
    *s = *Begin;
    
    sp = sbeg = s;
    Boundary->lsgl = dlhcons(Boundary->lsgl,s);
   
 }

  /**** there exists closing segments ****/

  else{
    Begin = (Sgl *)(Boundary->lsgl->el);
    sp = Begin;
    sbeg = sp; /* ?? */
  }
  Previous = Begin;

  Current = NULL;
  
  while(Current != Begin && Previous != NULL){
    
    /***** if there are extra closing segments ***/
    /***** Looking in the set of closing segment + map segment *****/
    
    if(ClosingSegments){
      if((Current = SearchNext(Previous,Boundary)) != NULL){
	Current->sfirst = sp;
	sp->slast = Current;
	Current->typf = sp->typl = 'V';
	sp = Current;
      }
      
      else if((Current = SearchNext(Previous,Map)) != NULL)
	{
	  s = Sgl_cree();
	  *s = *Current;
	  Boundary->lsgl = dlhcons(Boundary->lsgl,s);
	  s->sfirst = sp;
	  sp->slast = s;
	  s->typf = sp->typl = 'V';
	  sp = s;
	}
    }
    /********* if no closing segments: ********/
    /********Looking in the map only *********/
    
    else if((Current = SearchNext(Previous,Map)) != NULL)
      {
	s = Sgl_cree();
	*s = *Current;
	Boundary->lsgl = dlhcons(Boundary->lsgl,s);
	s->sfirst = sp;
	sp->slast = s;
	
	s->typf = sp->typl = 'V';
	sp = s;
	if(Current == Begin){
	  s->slast = sbeg;
	  sbeg->sfirst = s;
	  s->typl = sbeg->typf = 'V';
	}
      }
    Previous = Current;
  }
  
  if(Previous == NULL){
    printf("\n");
    printf("WARNING: Problem in linking the boundary;\n");
    printf("Boundary ignored: it will probably be harder\n");
    printf("to disambiguate ambiguous locations\n");
    printf("\n");
    /* freeing the boundary structure */
    
    Imgl_free(Boundary);
    Boundary = NULL;
    
  }
  /** checking the boundary is clockwise **/
  else if( poly_orientation(Boundary) >= 0){
    
    printf("\n");
    printf("WARNING: The boundary is not a clockwise polygon;\n");
    printf("Boundary ignored: it will probably be harder\n");
    printf("to disambiguate ambiguous locations\n");
    printf("\n");
    /* freeing the boundary structure */
    
    Imgl_free(Boundary);
    Boundary = NULL;
    
  }
	
  else{
    printf("\n");
    printf("Boundary: image number %ld\n",Bound_ind);
    printf("\n");
  }
}


static Sgl *SearchNext(Sgl *s,igla *ima)
{
  PDBL *ls,*l;
  Sgl *sn;

  if(ima == NULL)
    return(NULL);

  ls = ima->lsgl;
  
  for(l  = ls; l != NULL; l = dls(l)){
    sn = (Sgl  *)dlel(l);
    if(fabs(s->last.x - sn->first.x) < 1e-7 && 
       fabs(s->last.y - sn->first.y) < 1e-7)
      return(sn);
  }

  return(NULL);
}

/******* looks for the rightmost segment in ima
****** consequently, the returned segment must belong to 
****** the external boudary, if any */

static Sgl *ExternSegment(igla *ima)
{
  double x;
  PDBL *l;
  Sgl *s,*s_right = NULL;

  s  = (Sgl *)ima->lsgl->el;
  x = s->first.x;
  s_right = s;

  for(l = dls(ima->lsgl); l != NULL; l = dls(l)){
    s = (Sgl *)l->el;
    if( s->first.x > x){
      x = s->first.x;
      s_right = s;
    }
  }

  return(s_right);
}


/****************
  functions for setting parameters
****************/

#ifdef UNUSED
/*********
  setting the approximate angle if known.
  The Possible deviation is understood as 

  possible angle = mean plus or minus gap.

  Consequently, gap is between 0 and 180 degrees.
  NOTE: 180 degrees is equivalent to no restriction.

  Values are in degrees.
***********/

void Set_angle_restriction(void)
{
  double mean,gap;

  printf("Give aproximate steering angle\n");
  scanf("%lf",&mean);
  printf("Possible deviation:\n");
  scanf("%lf",&gap);
  Set_loc_angle(mean,gap);
}
#endif /* UNUSED */

/**************
  release the angle restriction
***************/

static void Unset_angle_restriction(void)
{
  Unset_loc_angle();
}
      
/**********
  define the accuracy on the extremity points 
  of the map segments. The accuracy is understood
  uniform and with no correlation between x and y
  coordinates.

  If there is already a map in memory, recomputes the map
  segment uncertainties on the supporting line
  parameters.

  The new value is kept as the default as long as the
  program is running.

  The default at the begining is 0.5
***********/

void Set_map_unaccuracy(double unc)
{
  Sgl *s;
  PDBL *l;

  unc_map = unc;
  /*
     if there is no map, returns and keep the new value
     as the new default value. Next time a map is read,
     the new value will be used.
  */
  if(Map == NULL)
    return;
  
  /***
    changing the map segment uncertainty 
  ***/

  for(l = Map->lsgl; l != NULL; l = dls(l)){
    
    s = (Sgl *)dlel(l);
   
    /* 
     Changing  the variance matrix on extremity points.
     The uncertainty is assumed uniform and
     with no correlation between x and y.
     Can be replaced by a more sophisticated function if needed.
     So far, works well.
    */
  
    s->first.vx = s->first.vy = unc_map*unc_map/12.;
    s->first.vxy = 0.;
    s->last.vx = s->last.vy = unc_map*unc_map/12.;
    s->last.vxy = 0.;
    
    /** 
      recomputing the variance on the supporting line
      parameters(las_unc.c) 
      **/
    
    initsg_map(&(s->first),&(s->last),s,unc_map);
  }
}



/* Mainly for debug purpose */
PDBL *GetListValidHypo()
{
  return ListValidHypo;
}


void segkitSetAspectDistMax(double distmax)
{
  segkitAspectDistMax = distmax;
}

void segkitGetAspectDistMax(double *distmax)
{
  *distmax = segkitAspectDistMax;
}
