#include <stdio.h>
#include <math.h>

#include <dbliste.h>
#include <lasdef.h>

#include "localize.h"
#include "print_results_p.h"

/* hyp->length_matched n'est dispos qu'apres TransformListHyp */
#define GAUSS_99 3.0 

/**
 **  print_hyp_list  -  
 **
 **  (hyp->length_matched n'est dispos qu'apres TransformListHyp)
 **/

void print_hyp_list (FILE *out,
		     PDBL *lhyp, Hyp *best, 
		     int lengthFlag, 
		     int printBestSegs, int printAllSegs)

{
  PDBL *l;
  Hyp *hyp;
  int nm;
  double ux, uy, ut;
  int i=0;

/*   l_end = dlfin(lhyp); */


/*   for(l = l_end; l != NULL; l = dlp(l)){ */
  for(l = lhyp; l != NULL; l = dls(l)){
    hyp = (Hyp *)dlel(l);

    if (best == hyp)
      fprintf (out, "BEST MATCHING :\n");
    
    nm = hyp->nmatched;
    if (nm == 0)
      continue;

    fprintf (out, "%-2d pos  (%-5.2f %-5.2f %-5.2f) nm %-2d lm %-.2g\n",
	     i++,
	     hyp->Tr[0], hyp->Tr[1], hyp->Tr[2], 
	     nm, hyp->length_matched);
    
    
    ux = sqrt(hyp->Vtr[0]) * GAUSS_99;
    uy = sqrt(hyp->Vtr[2]) * GAUSS_99;
    ut = sqrt(hyp->Vtr[5]) * GAUSS_99;
    
    fprintf (out, "   vpos (%.3f %.3f %.3f)\n",
	     ux, uy, ut);
    
    if (printBestSegs || printAllSegs) {
      if (printAllSegs || best == hyp) {
	print_hyp_segs(out, hyp->lsm, hyp->Tr, hyp->Vtr);
      }
    }
  }
}

/**
 **  print_pre_hyp_list  -
 **
 **  (hyp->length_matched n'est dispos qu'apres TransformListHyp)
 **/

void print_pre_hyp_list (FILE *out, PDBL *lprehyp, int showsegs)
{
  PDBL *l;
  Prehyp *prehyp;
  int i=0;

  l = lprehyp;

  do {
    prehyp = Prhel(l);

    fprintf (out, "   %d ang %5.3f vang %5.4f  nm %ld\n",
	    i++, prehyp->ang, prehyp->vang, prehyp->nmatched);

    if (showsegs) {
      print_hyp_segs_num(out, prehyp->lsm, NULL, NULL);
    }

    l = dls(l);

  } while(l != NULL && l != lprehyp);

}

/**
 **  print_hyp_segs  - Affiche les segments utiliss
 **
 **  (hyp->length_matched n'est dispos qu'apres TransformListHyp)
 **/

void print_hyp_segs(FILE *out, PDBL *lsm, double tr[4], double vtr[6])
{

  Smatch *match; /* 1 matching */
  PDBL *l,*l_end;
  Sgl *sp;  /* Le segment "percu" du matching */
  Sgl *sm;  /* Le segment "model" du matching */
  double eca; /* Ecart angulaire entre sgp et sgm */
  int i=0;

  /** the initial pairs of segments is at the end 
      of the list **/

  l_end = dlfin(lsm);


  for(l = l_end; l != NULL; l = dlp(l)){
    match = (Smatch *)(l->el);
    sm = match->sgm;
    sp = match->sgp;
    eca = match->eca;
    fprintf (out, "  %2d", i++);
    fprintf (out, "    m %3ld : (%6.3f %6.3f) (%6.3f %6.3f)   vo %.3f",
	    sm->ifirst, sm->first.x, sm->first.y, sm->last.x, sm->last.y, sm->vo);
    fprintf (out, " ang %.3f gpe %d\n", eca, match->numgp);
    fprintf (out, "        p %3ld : (%6.3f %6.3f) (%6.3f %6.3f)   vo %.3f\n",
	    sp->ifirst, sp->first.x, sp->first.y, sp->last.x, sp->last.y, sp->vo);
  } 
  
}

/**
 **  print_hyp_segs_num  - Affiche les segments utiliss
 **
 **  (hyp->length_matched n'est dispos qu'apres TransformListHyp)
 **/

void print_hyp_segs_num(FILE *out, PDBL *lsm, double tr[4], double vtr[6])
{

  Smatch *match; /* 1 matching */
  PDBL *l,*l_end;
  Sgl *sp;  /* Le segment "percu" du matching */
  Sgl *sm;  /* Le segment "model" du matching */
  double eca; /* Ecart angulaire entre sgp et sgm */
  int i=0;

  /** the initial pairs of segments is at the end 
      of the list **/

  l_end = dlfin(lsm);


  for(l = l_end; l != NULL; l = dlp(l)){
    match = (Smatch *)(l->el);
    sm = match->sgm;
    sp = match->sgp;
    eca = match->eca;
    fprintf (out, "  %2d ", i++);
    fprintf (out, " ang %.3f  gpe %d", eca, match->numgp);
    fprintf (out, "    m %3ld  vo %.3f",
	    sm->ifirst, sm->vo);
    fprintf (out, "    p %3ld  vo %.3f\n",
	    sp->ifirst, sp->vo);
  } 
  
}
